import subprocess
from pathlib import Path

import yaml

from playbacker.core.setlist import FileSetlist
from playbacker.core.settings import load_settings
from playbacker.core.song import load_songs


def format_yaml_files(*files: str | Path):
    subprocess.check_call(
        ["npx", "prettier", "--write", *files],
        stdout=subprocess.DEVNULL,
    )


def validate_config_schema(config_file: Path):
    with config_file.open() as f:
        content = yaml.safe_load(f)
    load_settings(content, "default")


def get_unknown_songs(songs_file: Path, setlists_dir: Path):
    with songs_file.open() as f:
        content = yaml.safe_load(f)
    songs = [s.name.casefold() for s in load_songs(content)]
    res: dict[Path, list[str]] = {}

    for setlist_path in setlists_dir.glob("*.yaml"):
        with setlist_path.open() as f:
            content = yaml.safe_load(f)
        cur_songs = FileSetlist(__root__=content).__root__

        if cur_unknown_songs := list(
            filter(lambda s: s.casefold() not in songs, cur_songs)
        ):
            res[setlist_path] = cur_unknown_songs

    return res



from pathlib import Path
from typing import Any, TypedDict

import sounddevice
from pydantic import BaseModel, Field

from playbacker.core.audiofile import AudioFile
from playbacker.core.tracks.countdown import CountdownSounds
from playbacker.core.tracks.metronome import MetronomeSounds


class _Sounds(BaseModel):
    metronome: MetronomeSounds
    countdown: CountdownSounds


class _ChannelMap(BaseModel):
    metronome: list[int]
    guide: list[int]
    multitrack: list[int]


class Settings(BaseModel, frozen=True):
    device: str | None
    sample_rate: int
    channel_map: _ChannelMap
    sounds: _Sounds
    channel_limit: int


class _Device(BaseModel):
    name: str | None
    pretty_name: str
    sample_rate: int
    channel_map: _ChannelMap


class _MetronomePaths(BaseModel):
    accent: Path
    tick_1_4: Path = Field(alias="1/4")
    tick_1_8: Path = Field(alias="1/8")
    tick_1_16: Path = Field(alias="1/16")


class _CountdownPaths(BaseModel):
    count_1: Path
    count_2: Path
    count_3: Path
    count_4: Path


class _SoundPaths(BaseModel):
    metronome: _MetronomePaths
    countdown: _CountdownPaths


class _FileSettings(BaseModel):
    devices: list[_Device]
    sounds: _SoundPaths


class _DeviceProps(TypedDict):
    max_output_channels: int


def _find_device_in_settings(name: str, devices: list[_Device]) -> _Device:
    for device in devices:
        if device.pretty_name == name:
            return device

    raise Exception(f"Device {name} not found in config")


def _get_device_props(device: _Device) -> _DeviceProps:
    return sounddevice.query_devices(  # pyright: ignore
        device=device.name, kind="output"
    )


def _get_channel_limit(device: _Device, device_props: _DeviceProps):
    map = device.channel_map
    largest_channel_no = max(map.metronome + map.guide + map.multitrack)

    limit = device_props["max_output_channels"]

    if largest_channel_no > limit:
        raise RuntimeError(
            f"Device {device.name} has only {limit}"
            + f" outputs, requested {largest_channel_no}."
        )

    return limit


def _convert_file_settings(settings: _FileSettings, device_name: str) -> Settings:
    device = _find_device_in_settings(name=device_name, devices=settings.devices)
    props = _get_device_props(device)
    channel_limit = _get_channel_limit(device, props)
    metronome = settings.sounds.metronome
    countdown = settings.sounds.countdown

    return Settings(
        device=device.name,
        sample_rate=device.sample_rate,
        channel_map=device.channel_map,
        sounds=_Sounds(
            metronome=MetronomeSounds(
                accent=AudioFile(metronome.accent, device.sample_rate),
                tick_1_4=AudioFile(metronome.tick_1_4, device.sample_rate),
                tick_1_8=AudioFile(metronome.tick_1_8, device.sample_rate),
                tick_1_16=AudioFile(metronome.tick_1_16, device.sample_rate),
            ),
            countdown=CountdownSounds(
                count_1=AudioFile(countdown.count_1, device.sample_rate),
                count_2=AudioFile(countdown.count_2, device.sample_rate),
                count_3=AudioFile(countdown.count_3, device.sample_rate),
                count_4=AudioFile(countdown.count_4, device.sample_rate),
            ),
        ),
        channel_limit=channel_limit,
    )


def load_settings(content: Any, device_name: str) -> Settings:
    return _convert_file_settings(_FileSettings(**content), device_name)



from pathlib import Path

from pydantic import BaseModel


class Config(BaseModel):
    config_dir_path: Path
    device: str


def get_config_file_path(config_dir_path: Path):
    path = config_dir_path / "config.yaml"
    assert path.exists()
    return path


def get_songs_file_path(config_dir_path: Path):
    path = config_dir_path / "songs.yaml"
    assert path.exists()
    return path


def get_setlists_dir_path(config_dir_path: Path):
    path = config_dir_path / "setlists"
    assert path.exists()
    return path



from typing import Literal

from pydantic import BaseModel

TimeSignature = Literal["4/4", "6/8", "12/4"]
Duration = Literal["1/4", "1/8", "1/16"]


class Tempo(BaseModel, frozen=True):
    bpm: float
    time_signature: TimeSignature
    duration: Duration

    @property  # TODO: Cached
    def lag(self) -> float:
        return 60 / self.bpm / 4

    @property
    def beats_per_bar(self) -> int:
        return int(self.time_signature[0])

    def get_start_of_bar(self, position: int) -> int:
        divider = self.beats_per_bar * 4
        return position - position % divider

    def __str__(self) -> str:
        pretty_bpm = int(self.bpm) if self.bpm.is_integer() else self.bpm
        return f"{pretty_bpm} bpm, {self.time_signature}, {self.duration}"



from dataclasses import dataclass, field
from typing import Literal, NamedTuple

from playbacker.core.audiofile import AudioArray, AudioFile
from playbacker.core.tempo import TimeSignature
from playbacker.core.track import SoundTrack


class CountdownSounds(NamedTuple):
    count_1: AudioFile
    count_2: AudioFile
    count_3: AudioFile
    count_4: AudioFile


_Instruction = Literal[1, 2, 3, 4, None]


@dataclass
class CountdownTrack(SoundTrack[CountdownSounds]):
    instruction_to_sound: dict[_Instruction, AudioFile] = field(init=False, repr=False)

    def __post_init__(self) -> None:
        super().__post_init__()
        self.instruction_to_sound = {
            1: self.sounds.count_1,
            2: self.sounds.count_2,
            3: self.sounds.count_3,
            4: self.sounds.count_4,
        }

    def get_sound(self) -> AudioArray | None:
        for end, instruction in countdown_schemes[self.shared.tempo.time_signature]:
            if self.shared.position < end:
                return self.instruction_to_sound[instruction].data

            if self.shared.position == end:
                self.current_frame = 0
                return None


class _Entry(NamedTuple):
    end: int
    instruction: _Instruction


countdown_schemes: dict[TimeSignature, list[_Entry]] = {
    "4/4": [
        _Entry(end=7, instruction=1),
        _Entry(end=15, instruction=2),
        _Entry(end=19, instruction=1),
        _Entry(end=23, instruction=2),
        _Entry(end=27, instruction=3),
        _Entry(end=31, instruction=4),
    ],
    "6/8": [
        _Entry(end=5, instruction=1),
        _Entry(end=11, instruction=2),
        _Entry(end=14, instruction=1),
        _Entry(end=17, instruction=2),
        _Entry(end=20, instruction=3),
        _Entry(end=23, instruction=4),
    ],
    "12/4": [
        _Entry(end=5, instruction=1),
        _Entry(end=11, instruction=2),
        _Entry(end=14, instruction=1),
        _Entry(end=17, instruction=2),
        _Entry(end=20, instruction=3),
        _Entry(end=23, instruction=4),
    ],
}



import time
from collections.abc import Callable
from dataclasses import dataclass, field
from threading import Event, Thread
from typing import Any, Protocol

import sounddevice

from playbacker.core.audiofile import AudioArray

SoundGetter = Callable[[int], AudioArray | None]


@dataclass
class Stream(Protocol):
    sound_getter: SoundGetter = field(repr=False)
    # Whether stream is already created in thread
    ready: Event = field(default_factory=Event, init=False, repr=False)
    sample_rate: int = field(repr=False)

    def destroy(self) -> None:  # pragma: no cover
        ...


def convert_channel_map_to_coreaudio_format(
    map: list[int], channel_limit: int
) -> list[int]:
    channel_map: dict[int, int] = {0: -1}
    prev_idx = -1

    for _i in range(channel_limit):
        idx = _i + 1

        if idx in map:
            channel_map[idx] = prev_idx + 1
            prev_idx += 1
        else:
            channel_map[idx] = -1

    return list(channel_map.values())


def allocate_data_to_channels(
    data: AudioArray, channel_map: list[int], channel_limit: int
) -> dict[int, AudioArray]:
    channel_to_data: dict[int, AudioArray] = {}
    prev_idx = 0

    for _i in range(channel_limit):
        if prev_idx >= data.shape[1]:  # no more channels
            break

        idx = _i + 1

        if idx in channel_map:
            channel_to_data[idx] = data[:, prev_idx : prev_idx + 1]
            prev_idx += 1

    return channel_to_data


@dataclass
class SounddeviceStream(Stream):
    """Wrapper around sounddevice.OutputStream"""

    channel_map: list[int]
    channel_limit: int = field(repr=False)
    device_name: str | None = field(repr=False)
    stream: sounddevice.OutputStream = field(init=False, repr=False)

    def _init_stream(self) -> None:
        map = convert_channel_map_to_coreaudio_format(
            self.channel_map, self.channel_limit
        )
        extra = sounddevice.CoreAudioSettings(channel_map=map)
        self.stream = sounddevice.OutputStream(
            samplerate=self.sample_rate,
            device=self.device_name,
            extra_settings=extra,
            callback=self._callback,
        )
        self.stream.start()
        self.ready.set()

    def __post_init__(self) -> None:
        Thread(target=self._init_stream, daemon=True).start()

    def _destroy_stream(self) -> None:
        time.sleep(0.4)
        self.stream.stop()
        self.stream.close()

    def destroy(self) -> None:
        Thread(target=self._destroy_stream).start()

    def _fill_outdata(self, outdata: AudioArray, data: AudioArray) -> None:
        map = allocate_data_to_channels(data, self.channel_map, self.channel_limit)
        size = len(data)
        for idx, data_for_channel in map.items():
            outdata[:size, idx - 1 : idx] = data_for_channel

    def _callback(
        self,
        outdata: AudioArray,
        frames: int,
        time: Any,
        status: sounddevice.CallbackFlags,
    ) -> None:
        if (data := self.sound_getter(frames)) is None:
            outdata.fill(0)
        else:
            self._fill_outdata(outdata, data)



from pathlib import Path
from typing import Any

from pydantic import BaseModel, Field

from playbacker.core.tempo import Tempo


class _SongBase(BaseModel, frozen=True):
    artist: str | None
    tempo: Tempo


class Song(_SongBase, frozen=True):
    name: str


class _TrackPaths(BaseModel):
    multitrack: Path | None = None
    guide: Path | None = None


class _FileSong(_SongBase, frozen=True):
    tracks: _TrackPaths = Field(default_factory=_TrackPaths)


class _FileSongs(BaseModel):
    __root__: dict[str, _FileSong]


def _convert_file_song(name: str, song: _FileSong) -> Song:
    return Song(
        name=name,
        artist=song.artist,
        tempo=song.tempo,
    )


def load_songs(content: Any) -> list[Song]:
    songs = _FileSongs(__root__=content)
    return [
        _convert_file_song(name=name, song=song)
        for name, song in songs.__root__.items()
    ]



import asyncio
import subprocess
from pathlib import Path
from typing import Any

import watchfiles
import yaml
from pydantic import BaseModel
from starlite import (
    Controller,
    CORSConfig,
    Dependency,
    NotFoundException,
    Provide,
    Starlite,
    StaticFilesConfig,
    WebSocket,
    post,
    websocket,
)

from playbacker.config import (
    Config,
    get_config_file_path,
    get_setlists_dir_path,
    get_songs_file_path,
)
from playbacker.core.playback import Playback
from playbacker.core.player import Player
from playbacker.core.setlist import (
    NoSongInStorageError,
    Setlist,
    load_setlist,
    prettify_setlist_stem,
)
from playbacker.core.settings import load_settings
from playbacker.core.song import load_songs
from playbacker.core.tempo import Tempo


def get_setlist_path_from_pretty_name(name: str, setlists_dir_path: Path) -> Path:
    map = {prettify_setlist_stem(f.stem): f for f in setlists_dir_path.glob("*.yaml")}
    if path := map.get(name):
        return path
    raise NotFoundException(detail="no setlist with this name")


class PlayerState(BaseModel):
    playing: bool
    guide_enabled: bool

    @classmethod
    def make(cls, player: Player):
        return cls(playing=player.playing, guide_enabled=player.guide_enabled)


class MainController(Controller):
    @post("/get_setlists")
    def get_setlists(self, setlists_dir: Path) -> list[str]:
        stems = [prettify_setlist_stem(f.stem) for f in setlists_dir.glob("*.yaml")]
        stems.sort(reverse=True)
        return stems

    @post("/get_setlist")
    def get_setlist(self, name: str, setlists_dir: Path, songs_file: Path) -> Setlist:
        path = get_setlist_path_from_pretty_name(name, setlists_dir)

        with songs_file.open() as f:
            songs = load_songs(content=yaml.safe_load(f))

        with path.open() as f:
            content = yaml.safe_load(f)

        try:
            return load_setlist(name=name, content=content, songs=songs)
        except NoSongInStorageError as err:
            raise NotFoundException(detail=err.message)

    @post("/toggle_playing")
    def toggle_playing(
        self, data: Tempo, player: Player = Dependency(skip_validation=True)
    ) -> PlayerState:
        if player.playing:
            player.pause()
        else:
            player.play(data)
        return PlayerState.make(player)

    @post("/toggle_guide_enabled")
    def toggle_guide_enabled(
        self, player: Player = Dependency(skip_validation=True)
    ) -> PlayerState:
        player.guide_enabled = not player.guide_enabled
        return PlayerState.make(player)

    @post("/prepare_for_switch")
    def prepare_for_switch(
        self, player: Player = Dependency(skip_validation=True)
    ) -> PlayerState:
        player.prepare_for_switch()
        return PlayerState.make(player)

    @post("/reset")
    def reset(self, player: Player = Dependency(skip_validation=True)) -> PlayerState:
        player.reset()
        return PlayerState.make(player)


@websocket("/watch")
async def watch_route(
    socket: WebSocket[Any, Any],
    current_setlist: str,
    setlists_dir: Path,
    songs_file: Path,
) -> None:
    await socket.accept()
    setlist_path = get_setlist_path_from_pretty_name(current_setlist, setlists_dir)

    async def run():
        async for changes in watchfiles.awatch(  # pyright: ignore[reportUnknownMemberType]
            songs_file, setlists_dir
        ):
            for _, path in changes:
                path = Path(path)
                if path == setlist_path or path == songs_file:
                    await socket.send_text("current_setlist")
                else:
                    await socket.send_text("setlists")

    task = asyncio.create_task(run())
    try:
        await socket.receive()
    finally:
        task.cancel()


def get_app(config: Config):
    setlists_dir = get_setlists_dir_path(config.config_dir_path)
    songs_file = get_songs_file_path(config.config_dir_path)

    with get_config_file_path(config.config_dir_path).open() as f:
        content = yaml.safe_load(f)
    player = Player(Playback(load_settings(content=content, device_name=config.device)))

    frontend = Path(__file__).parent / "dist"
    with_frontend = frontend.exists()
    if with_frontend:
        static_files_config = StaticFilesConfig(
            directories=[frontend], path="/", html_mode=True
        )
    else:
        static_files_config = None

    def open_browser():
        if with_frontend:
            subprocess.check_call(("open", "http://127.0.0.1:8000"))

    return Starlite(
        route_handlers=[MainController, watch_route],
        dependencies={
            "setlists_dir": Provide(lambda: setlists_dir),
            "songs_file": Provide(lambda: songs_file),
            "player": Provide(lambda: player),
        },
        on_startup=[open_browser],
        on_shutdown=[player.stop],
        cors_config=CORSConfig(),
        static_files_config=static_files_config,
    )



from collections.abc import Iterator, Sequence
from contextlib import contextmanager
from dataclasses import dataclass, field
from typing import Generic, NamedTuple, Protocol, TypeVar

from playbacker.core.clock import Clock
from playbacker.core.settings import Settings
from playbacker.core.stream import SounddeviceStream
from playbacker.core.tempo import Tempo
from playbacker.core.track import Shared, StreamBuilder, Track
from playbacker.core.tracks.countdown import CountdownTrack
from playbacker.core.tracks.metronome import MetronomeTrack

_Tracks = TypeVar("_Tracks", bound=Sequence[Track])


@dataclass
class BasePlayback(Generic[_Tracks], Protocol):
    """Playback that manages clock and tracks."""

    clock: Clock = field(init=False)
    shared: Shared = field(default_factory=Shared, init=False)
    tracks: _Tracks = field(init=False)

    def __post_init__(self) -> None:
        self.clock = Clock(callback=self.clock_callback)
        self.tracks = self.get_tracks()

    def clock_callback(self) -> None:
        self.shared.position += 1

        for track in self.tracks:
            track.tick()

    def get_tracks(self) -> _Tracks:
        ...

    def resume(self) -> None:
        self.shared.position = self.shared.tempo.get_start_of_bar(self.shared.position)

        for track in self.tracks:
            track.resume()

        self.clock.start()

    def pause(self) -> None:
        for track in self.tracks:
            track.pause()

        self.clock.pause()

    def destroy(self) -> None:
        """Graceful shutdown. In this order (almost) no glitches in the end."""
        self.pause()

        self.clock.destroy()

        for track in self.tracks:
            track.destroy()

    def start(self, *args: ..., **kwargs: ...) -> None:
        ...

    @contextmanager
    def starting_ctx(self, tempo: Tempo) -> Iterator[None]:
        self.shared.configuring = True
        self.shared.position = 0
        self.shared.tempo = tempo
        yield
        self.shared.configuring = False
        self.clock.lag = tempo.lag
        self.clock.start()


class DefaultTracks(NamedTuple):
    metronome: MetronomeTrack
    countdown: CountdownTrack


@dataclass
class Playback(BasePlayback[DefaultTracks]):
    """Playback that deals with settings and actual tracks"""

    settings: Settings = field(repr=False)

    def get_tracks(self) -> DefaultTracks:
        def create_builder(channel_map: list[int]) -> StreamBuilder:
            return lambda g: SounddeviceStream(
                sound_getter=g,
                sample_rate=self.settings.sample_rate,
                channel_map=channel_map,
                channel_limit=self.settings.channel_limit,
                device_name=self.settings.device,
            )

        map = self.settings.channel_map

        return DefaultTracks(
            metronome=MetronomeTrack(
                shared=self.shared,
                sounds=self.settings.sounds.metronome,
                stream_builder=create_builder(map.metronome),
            ),
            countdown=CountdownTrack(
                shared=self.shared,
                sounds=self.settings.sounds.countdown,
                stream_builder=create_builder(map.guide),
            ),
        )

    def start(self, tempo: Tempo) -> None:
        with self.starting_ctx(tempo):
            self.tracks.metronome.start()
            self.tracks.countdown.start()



from dataclasses import dataclass, field
from functools import cached_property
from pathlib import Path
from typing import Any, cast

import numpy
import soundfile
import soxr

AudioArray = numpy.ndarray[Any, numpy.dtype[Any]]


@dataclass
class AudioFile:
    path: Path
    sample_rate: int = field(repr=False)

    @cached_property
    def data(self) -> AudioArray:
        """Read audio file and convert to required sample rate."""
        data, in_rate = cast(
            tuple[AudioArray, int],
            soundfile.read(self.path),  # pyright: ignore[reportUnknownMemberType]
        )

        if in_rate == self.sample_rate:
            return data

        return cast(
            AudioArray,
            soxr.resample(  # pyright: ignore[reportUnknownMemberType]
                data, in_rate, self.sample_rate
            ),
        )



from collections.abc import Callable, Iterable
from dataclasses import dataclass, field
from typing import Generic, Protocol, TypeVar

from playbacker.core.audiofile import AudioArray, AudioFile
from playbacker.core.stream import SoundGetter, Stream
from playbacker.core.tempo import Tempo


@dataclass
class Shared:
    tempo: Tempo = field(init=False)
    configuring: bool = True
    position: int = 0


@dataclass
class Track(Protocol):
    shared: Shared = field()
    paused: bool = field(default=False, init=False)

    def tick(self) -> None:  # pragma: no cover
        pass

    def destroy(self) -> None:  # pragma: no cover
        ...

    def resume(self) -> None:  # pragma: no cover
        ...

    def pause(self) -> None:  # pragma: no cover
        ...


def trim_audio_array(
    data: AudioArray, current_frame: int, required_frames: int
) -> AudioArray:
    chunk_size = min(len(data) - current_frame, required_frames)
    trimmed = data[current_frame : current_frame + chunk_size]
    return trimmed


_Sounds = TypeVar("_Sounds", bound=Iterable[AudioFile | None])
StreamBuilder = Callable[[SoundGetter], Stream]


@dataclass
class SoundTrack(Track, Generic[_Sounds], Protocol):
    paused: bool = field(default=False, init=False)
    stream_builder: StreamBuilder = field()
    stream: Stream = field(init=False)
    current_frame: int = field(default=0, init=False)
    sounds: _Sounds = field(repr=False)
    enabled: bool = field(default=True, init=False)

    def __post_init__(self) -> None:
        self.stream = self.stream_builder(self.callback)

    def get_sound(self) -> AudioArray | None:  # pragma: no cover
        ...

    def _should_be_silent(self) -> bool:
        return not self.enabled or self.paused or self.shared.configuring

    def callback(self, frames: int) -> AudioArray | None:
        if self._should_be_silent() or (data := self.get_sound()) is None:
            return

        trimmed = trim_audio_array(data, self.current_frame, frames)
        self.current_frame += len(trimmed)
        return trimmed

    def resume(self) -> None:
        self.paused = False

    def pause(self) -> None:
        self.paused = True

    def destroy(self) -> None:
        self.stream.destroy()

    def _preload_audiofiles(self) -> None:
        for sound in self.sounds:
            if sound:
                sound.data

    def start(self, *, sounds: _Sounds | None = None) -> None:
        self.current_frame = 0
        self.stream.ready.wait()
        self.resume()

        if sounds:
            self.sounds = sounds

        self._preload_audiofiles()



from typing import Any

from pydantic import BaseModel

from playbacker.core.song import Song


class Setlist(BaseModel):
    name: str
    songs: list[Song]


class NoSongInStorageError(Exception):
    message: str

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__()


def _find_song_in_storage(name: str, storage: list[Song]):
    for song in storage:
        if song.name.casefold() == name.casefold():
            return song

    raise NoSongInStorageError(f'Song "{name}" is not present in storage')


class FileSetlist(BaseModel):
    __root__: list[str]


def load_setlist(name: str, content: Any, songs: list[Song]) -> Setlist:
    fsetlist = FileSetlist(__root__=content)
    selected_songs = [_find_song_in_storage(n, songs) for n in fsetlist.__root__]
    return Setlist(name=name, songs=selected_songs)


def prettify_setlist_stem(stem: str) -> str:
    return " ".join(w.capitalize() for w in stem.split())



import time
from collections.abc import Callable
from dataclasses import dataclass, field
from threading import Event, Thread
from typing import NoReturn


@dataclass
class Clock:
    callback: Callable[[], None] = field(repr=False)
    thread: Thread = field(init=False, repr=False)
    started: Event = field(default_factory=Event, init=False, repr=False)
    lag: float = field(init=False)
    previous_time: float = field(init=False, repr=False)

    def __post_init__(self) -> None:
        self.thread = Thread(daemon=True, target=self.run)
        self.thread.start()

    def start(self) -> None:
        self.previous_time = time.monotonic()
        self.started.set()

    def pause(self) -> None:
        self.started.clear()

    def destroy(self) -> None:
        self.thread.join(0)

    def _sleep(self):
        result = self.lag + self.previous_time - time.monotonic()
        sleep_for = result * 0.925 if result > 0 else 0
        time.sleep(sleep_for)

    def _tick(self) -> None:
        self._sleep()
        self.previous_time += self.lag
        self.callback()

    def _run_once(self) -> None:
        self.started.wait()
        while self.started.is_set():
            self._tick()

    def run(self) -> NoReturn:
        while True:
            self._run_once()



from dataclasses import dataclass, field
from typing import Literal, NamedTuple

from playbacker.core.audiofile import AudioArray, AudioFile
from playbacker.core.tempo import Duration, Tempo, TimeSignature
from playbacker.core.track import SoundTrack


class MetronomeSounds(NamedTuple):
    accent: AudioFile
    tick_1_4: AudioFile
    tick_1_8: AudioFile
    tick_1_16: AudioFile


_Instruction = Literal["accent", 4, 8, 16]


@dataclass
class MetronomeTrack(SoundTrack[MetronomeSounds]):
    instruction_to_sound: dict[_Instruction, AudioFile] = field(init=False, repr=False)

    def __post_init__(self) -> None:
        super().__post_init__()
        self.instruction_to_sound = {
            "accent": self.sounds.accent,
            4: self.sounds.tick_1_4,
            8: self.sounds.tick_1_8,
            16: self.sounds.tick_1_16,
        }

    def get_sound(self) -> AudioArray | None:
        if instruction := get_instruction(self.shared.tempo, self.shared.position):
            return self.instruction_to_sound[instruction].data

    def tick(self) -> None:
        self.current_frame = 0


class _Entry(NamedTuple):
    divider: int
    instruction: _Instruction


metronome_schemes: dict[TimeSignature, dict[Duration, list[_Entry]]] = {
    "4/4": {
        "1/4": [_Entry(divider=4, instruction=4)],
        "1/8": [
            _Entry(divider=16, instruction="accent"),
            _Entry(divider=4, instruction=4),
            _Entry(divider=2, instruction=8),
        ],
        "1/16": [
            _Entry(divider=4, instruction=4),
            _Entry(divider=2, instruction=8),
            _Entry(divider=1, instruction=16),
        ],
    },
    "6/8": {
        "1/4": [_Entry(divider=6, instruction=4)],
        "1/8": [_Entry(divider=6, instruction=4), _Entry(divider=2, instruction=8)],
        "1/16": [
            _Entry(divider=6, instruction=4),
            _Entry(divider=2, instruction=8),
            _Entry(divider=1, instruction=16),
        ],
    },
    "12/4": {
        "1/4": [_Entry(divider=12, instruction="accent")],
        "1/8": [
            _Entry(divider=12, instruction="accent"),
            _Entry(divider=2, instruction=8),
        ],
        "1/16": [
            _Entry(divider=12, instruction="accent"),
            _Entry(divider=2, instruction=8),
            _Entry(divider=1, instruction=16),
        ],
    },
}


def get_instruction(tempo: Tempo, position: int) -> _Instruction | None:
    for divider, instruction in metronome_schemes[tempo.time_signature][tempo.duration]:
        if position % divider == 0:
            return instruction



from pathlib import Path

import typer
import uvicorn
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich.table import Table

from playbacker.app import get_app
from playbacker.config import (
    Config,
    get_config_file_path,
    get_setlists_dir_path,
    get_songs_file_path,
)
from playbacker.core.validate import (
    format_yaml_files,
    get_unknown_songs,
    validate_config_schema,
)

default_config_dir_path = Path("~/.config/playbacker").expanduser()
config_opt = typer.Option(
    default_config_dir_path, exists=True, file_okay=False, dir_okay=True
)
default_device = "default"


def default_app():
    default_config = Config(
        config_dir_path=default_config_dir_path, device=default_device
    )
    return get_app(default_config)


cli = typer.Typer(add_completion=False)


@cli.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    device: str = default_device,
    config: Path = config_opt,
    reload: bool = False,
):
    if ctx.invoked_subcommand is not None:
        return

    if reload:
        if device != default_device:
            print(f"Setting device to {default_device!r} since you passed --reload.")
        if config != default_config_dir_path:
            print(
                f"Setting config to {default_config_dir_path} since you passed --reload."
            )
        uvicorn.run(  # pyright: ignore[reportUnknownMemberType]
            "playbacker.main:default_app",
            factory=True,
            reload=True,
        )

    else:
        app = get_app(Config(config_dir_path=config, device=device))
        uvicorn.run(app)  # pyright: ignore[reportUnknownMemberType]


@cli.command()
def validate(config: Path = config_opt):
    config_file = get_config_file_path(config)
    songs_file = get_songs_file_path(config)
    setlists_dir = get_setlists_dir_path(config)

    with Progress(
        SpinnerColumn(), TextColumn("[progress.description]{task.description}")
    ) as progress:
        progress.add_task(description="Checking unknown songs...")
        unknown_songs = get_unknown_songs(
            songs_file=songs_file, setlists_dir=setlists_dir
        )
        if unknown_songs:
            table = Table(show_header=False, show_lines=True)
            for setlist_path, songs in unknown_songs.items():
                songs_str = ",\n".join(f'"{song}"' for song in songs)
                table.add_row(str(setlist_path), songs_str)
            progress.stop()
            progress.print("There are unknown songs in setlists:", table)

        progress.add_task(description="Validating config schema...")
        validate_config_schema(config_file)

        progress.add_task(description="Formatting...")
        format_yaml_files(config_file, songs_file, f"{setlists_dir}/*.yaml")



from dataclasses import dataclass

from playbacker.core.playback import Playback
from playbacker.core.tempo import Tempo


@dataclass(init=False)
class Player:
    """Remembers current start-stop and pause-resume states."""

    playback: Playback
    tempo: Tempo | None = None
    started: bool = False
    playing: bool = False

    def __init__(self, playback: Playback) -> None:
        self.playback = playback

    def play(self, tempo: Tempo) -> None:
        if self.tempo == tempo:
            self.playback.resume()
        else:
            self.tempo = tempo
            self.playback.start(tempo=tempo)

        self.playing = True
        self.started = True

    def pause(self) -> None:
        self.playback.pause()
        self.playing = False

    def reset(self) -> None:
        self.pause()
        self.tempo = None

    def stop(self) -> None:
        self.playback.destroy()
        self.started = False

    def prepare_for_switch(self):
        if self.started:
            self.pause()
            self.guide_enabled = True

    @property
    def guide_enabled(self):
        return self.playback.tracks.countdown.enabled

    @guide_enabled.setter
    def guide_enabled(self, value: bool):
        self.playback.tracks.countdown.enabled = value

    def time(self) -> float:
        if self.started and self.tempo:
            return self.playback.shared.tempo.lag * self.playback.shared.position
        return 0
